import sys
import os
from os import remove, listdir, getcwd
from os.path import dirname, join
from shutil import copyfile
from getpass import getpass
from urlparse import urljoin
from lxml.html import document_fromstring
import pycurl
from chameleon.zpt.template import PageTemplateFile
import keyring

import sixfeetup.amalgam

# add in the directory into the python path
sys.path.append(getcwd())
import specbook_config
from specbook_config import config

TRUISMS = ['y', 'yes', 'true']


class Credential(object):

    def user_credentials(self, username=None):
        """
        """
        if username is None:
            username = os.getenv("USER", "")
            # if there is no USER env var...
            if not username:
                print "You must supply a username (-u)"
                sys.exit(1)
        self.username = username
        self.password = keyring.get_password('todo', username)
        if self.password is None:
            self.password = getpass("Trac Password: ")
            #if not keyring.supported():
            #    return
            save_password = raw_input(
                "Save this password in your keychain? [y/N] ")
            if save_password.lower() in TRUISMS:
                keyring.set_password('todo', self.username, self.password)


class TracPage(Credential):

    def __init__(self, base_url):
        self.contents = ''
        bu, qs = base_url.split('?')
        self.user_credentials()
        sn = bu % (self.username, self.password)
        self.fullurl = sn + '?' + qs

    def body_callback(self, buf):
        self.contents = self.contents + buf

    def get_tree(self):
        print "Fetching Trac page"

        c = pycurl.Curl()
        c.setopt(c.URL, self.fullurl)
        c.setopt(c.WRITEFUNCTION, self.body_callback)
        c.perform()
        c.close()

        tree = document_fromstring(self.contents)
        return tree


class Report(object):

    base_xp = '/*/body//table[@class="listing tickets"]/tbody'
    title = None
    milestone = None
    priority = None
    functionality = None
    hours_left = 0
    main_dev = None

    # Each phase in a report lifecycle
    wf = dict(ticket_id=None, summary=None, milestone=None,
              status=None, owner=None, remaining=None)
    sql = dict(hours=None, tickets=None, status=None, who=None)
    logging = dict(hours=None, tickets=None, status=None, who=None)
    alldata = dict(hours=None, tickets=None, status=None, who=None)
    calculations = dict(hours=None, tickets=None, status=None, who=None)
    filter = dict(hours=None, tickets=None, status=None, who=None)
    print_excel = dict(hours=None, tickets=None, status=None, who=None)
    tested = dict(hours=None, tickets=None, status=None, who=None)

    def __init__(self, reports_tree, subtasks_tree, report_id):

        self.report_id = report_id[1:]

        # Get the HTML row in the reports listing and grab the ticket ids for
        # all the "blocked by"
        xp1 = self.base_xp + '/tr//td[@class="id"][a/text()="%s"]/..'
        reports_row = reports_tree.xpath(xp1 % report_id)[0]

        # Look in the subtasks tree and make an dict of all the subtasks, keyed
        # by task id
        xp2 = self.base_xp + '/tr[td[@class="id"]]'
        alltasks = {}
        for subtask in subtasks_tree.xpath(xp2):
            stxp = subtask.xpath
            ticket_id = stxp('td[@class="id"]/a/text()')[0][1:].strip()
            summary = stxp('td[@class="summary"]/a/text()')[0].strip()
            milestone = stxp('td[@class="milestone"]/a/text()')[0].strip()
            status = stxp('td[@class="status"]/text()')[0].strip()
            owner = stxp('td[@class="owner"]/text()')[0].strip()
            remaining = stxp('td[@class="tt_remaining"]/text()')[0].strip()
            alltasks[ticket_id] = dict(summary=summary, milestone=milestone,
                                       status=status, owner=owner,
                                       remaining=remaining)

        subtask_ids = reports_row.xpath(
            'td/span/a["mt-assigned ticket"]/text()')
        for subtask_id in subtask_ids:
            subtask = alltasks.get(subtask_id[1:], None)
            if subtask is None:
                continue
            summary = subtask['summary']
            if summary.endswith('Finish wireframe'):
                self.wf = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Implement model into SA'):
                self.sql = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Implement auditing/logging'):
                self.logging = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Implement report for All Data'):
                self.alldata = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Implement report for Calculations'):
                self.calculations = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Hook up filters'):
                self.filter = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Wire up print and Excel views'):
                self.print_excel = subtask
                self.hours_left = self.hours_left + float(subtask['remaining'])
            elif summary.endswith('Test and report bugs'):
                self.tested = subtask
                # self.hours_left + float(subtask['remaining'])
                self.hours_left = 99.9


def list_reports():
    """Return sequence of Report instances"""

    tp1 = TracPage(config['reports_url'])
    reports_tree = tp1.get_tree()
    tp2 = TracPage(config['subtasks_url'])
    subtasks_tree = tp2.get_tree()

    xp = '/*/body//table[@class="listing tickets"]/tbody' + \
         '/tr//td[@class="id"]/a'
    reports = {}
    for report in reports_tree.xpath(xp):
        report_id = report.text
        rk = report.xpath('../../td[@class="summary"]/a/text()')[0]
        report_key = rk.split(' -')[0][8:].replace(' ', '')
        r = Report(reports_tree, subtasks_tree, report_id)
        reports[report_key] = r

    return reports


def main2():
    print 'Started'
    reports = list_reports()
    print 'Finished'


class WikiFetcher(Credential):

    url = urljoin(config['wiki_base_url'], config['hierarchy'][0]['trac_url'])

    def __init__(self):
        self.contents = ''
        self.user_credentials()
        self.fullurl = self.url % (self.username, self.password)

    def body_callback(self, buf):
        self.contents = self.contents + buf

    def extract_legends(self):
        print "Fetching wikipage"

        c = pycurl.Curl()
        c.setopt(c.URL, self.fullurl)
        c.setopt(c.WRITEFUNCTION, self.body_callback)
        c.perform()
        c.close()

        tree = document_fromstring(self.contents)

        # First the legends, then the "other specs"
        legends = {}
        xp = '//li/i[starts-with(text(), "{")]/text()'
        for legend in tree.xpath(xp):
            # The text of the note in the <li>
            # XXX This will fail if there is mixed content
            legend_body = legend.getparent().getparent().text.strip()

            # The text of the curly braces content
            legend_text = legend.strip().split('{')[1].split('}')[0]
            if ':' not in legend_text:
                # This isn't of the form {foo:99} so it isn't a legend, so skip
                continue
            screen_id, legend_id = legend_text.split(':')
            if screen_id in legends:
                legends[screen_id][legend_id] = legend_body
            else:
                legends[screen_id] = {legend_id: legend_body}

        # Now sort
        def sort_key(v):
            return int(v[0])
        for l in legends.keys():
            these_values = legends[l].items()
            new_values = sorted(these_values, key=sort_key)
            legends[l] = new_values

        # Now the other specs
        other_specs = {}
        xp = '//li/i[starts-with(text(), "{")]/text()'
        for other_spec in tree.xpath(xp):
            # The text of the note in the <li>
            # XXX This will fail if there is mixed content
            other_spec_body = other_spec.getparent().getparent().text.strip()

            # The text of the curly braces content
            other_spec_text = other_spec.strip().split('{')[1].split('}')[0]
            if ':' in other_spec_text:
                # This *is* of the form {foo:99} so it is a legend, so skip
                continue
            screen_id = other_spec_text
            if screen_id in other_specs:
                other_specs[screen_id].append(other_spec_body)
            else:
                other_specs[screen_id] = [other_spec_body, ]
        return legends, other_specs

    def update_modified(self, wf_jpg_fn):
        """Update the modification dates on wireframe data"""

        changes_info = {
            'some_screen_id': {
                'legends': {
                    '1': ['12/01', 'red'],
                    '2': ['12/02', 'black'],
                },
            }
        }

        return changes_info


def main():
    docsdir = dirname(specbook_config.__file__)
    module_dir = dirname(sixfeetup.amalgam.__file__)
    exported = join(docsdir, 'export')
    outputdir = join(docsdir, 'output')

    wf = WikiFetcher()
    legends, other_specs = wf.extract_legends()

    # Clear contents of the output direction and copy CSS into it
    for filename in listdir(outputdir):
        if '.svn' not in filename:
            remove(join(outputdir, filename))
    copyfile(join(module_dir, 'specbook.css'), join(outputdir, 'specbook.css'))

    # ZPT used for all screen wireframes
    screen_zpt = PageTemplateFile(join(module_dir, 'wrapper.pt'), None)

    # Generate cover pages for all the functionalities
    functionality_zpt = PageTemplateFile(
        join(module_dir, 'functionality.pt'), None)
    this_workstream = config['hierarchy'][0]
    functionalities = this_workstream['functionalities']
    for this_functionality in functionalities:
        fn = join(docsdir, 'output', this_functionality['url'] + '.html')
        f = open(fn, 'w')
        output = functionality_zpt(
            this_functionality=this_functionality,
            this_workstream=this_workstream,
            config=config)
        f.write(output)
        f.close()

        # Now do the screens for this functionality
        for ts in this_functionality['screens']:

            # Read contents of ISMAP file
            ismap = open(join(exported, ts['export_id'] + '.html')).read()

            # Make legend bullets
            this_screen_id = ts['export_id'].split('_')[0]
            these_legends = legends.get(this_screen_id, None)
            these_other_specs = other_specs.get(this_screen_id, None)

            # Copy JPG over
            wf_jpg_fn = join(exported, ts['export_id'] + '.jpg')
            copyfile(wf_jpg_fn,
                     join(docsdir, 'output', ts['export_id'] + '.jpg'))

            # Update the JSON database the tracks modified dates on wireframe
            # JPEGS, legends, and other specs
            changes_info = wf.update_modified(ts['export_id'])

            # Write output
            f = open(join(docsdir, 'output', ts['export_id'] + '.html'), 'w')
            output = screen_zpt(
                    these_legends=these_legends,
                    these_other_specs=these_other_specs,
                    ismap=ismap,
                    config=config,
                    this_screen=ts,
                    this_functionality=this_functionality,
                    this_workstream=this_workstream,
                        )
            f.write(output)
            f.close()


if __name__ == '__main__':
    main()
    print "Done"
